/* Find the polynomial regression of a data set.

   What needs to be passed:

   float **regflux - An array containing fluxes as flux[map][region]. A single region is allowable but the second dimension of the array must still be included.
   float *specindex - An empty array to return the spectral indices. Must be of format specindex[region]
   int numregions - The number of regions in the dataset.
   float freq1 - Frequency of maps 1
   float freq2 - Frequency of maps 2
   float *maxsi - Holder for the maximum spectral index
   float *minsi - Holder for the maximum spectral index
   int printindex - Print the spectral index to the terminal?


Basic usage. We assume imgnum is already set to be an integer number of data points (maps), flux is an array of fluxes by region, and map1 and maps 2 are set to the desired map values for flux[map] (normally min and max frequencies):

  float **alpha;
  alpha = calloc(imgnum, sizeof(float *));

  ...

  calcspecindex(flux, alpha, numregions, map1, map2, freq1, freq2, 1);

Plain values (not logs) should be passed. Multiple spectal indices can be obtained using a simple loop.
This script is designed to be combined with the spectra index mapping routine.

   If you have any question please email jeremy.harwood@physics.org and I will try and answer them as soon as I can
*/


int calcspecindex(float **regflux,  int **regionarray, float *specindex, int numregions, int imgnum, float *frequency, int printindex, float *minsi, float *maxsi) {

  // Declaring variables

  int i, j;
  double sumX, sumY, sumXX, sumXY;

  *maxsi = -1e32;
  *minsi = 1e32;

 
  // Determine the spectral indices between the maximum and minimum data points (maps)

  for (i = 1; i <=numregions; i++) {

      sumX = 0;
      sumY = 0;
      sumXX = 0;
      sumXY = 0;

    for (j = 0; j < imgnum; j++) {
      sumX += log(frequency[j]);
      sumY += log(regflux[j][i]);
      sumXX += log(frequency[j]) * log(frequency[j]);
      sumXY += log(frequency[j]) * log(regflux[j][i]);
    }



    // printf("sumX: %.2e sumY: %.2e sumXX: %.2e sumXY: %.2e\n",sumX,sumY,sumXX,sumXY);


    specindex[i] = -( ((imgnum * sumXY) - (sumX * sumY)) / ((imgnum * sumXX) - (sumX * sumX)) );

	// Find the max and min spectral indices as we go
	if (specindex[i] > *maxsi) {
	  *maxsi = specindex[i];
	}
	if (specindex[i] < *minsi) {
	  *minsi = specindex[i];
	}


    if (printindex == 1) {

      printf("Spectal Index of region %d: %.2f\n", i, specindex[i]);
    }
  }

  /*
  if (warningcount > 0) {
    fprintf(stderr,"\n*** WARNING: Regions had a flux less than the RMS %d times! This normally occurs when fixed regions are used and they are not a good match ***\n\n", warningcount);
  }
  */

    printf("Maximum Spectal Index: %.2f\n", *maxsi);
    printf("Minimum Spectal Index: %.2f\n", *minsi);

  return 0;

}



// Calculate the spec index using only 2 points e.g. for colour-colour plots 

int calcspecindexp2p(float **regflux, float *specindex, int numregions, int map1, int map2, float freq1, float freq2, int printindex, float *minsi, float *maxsi, float *rms) {

  // Declaring variables

  int i;
  int warningcount = 0;
  *maxsi = -1e32;
  *minsi = 1e32;

 
  // Determine the spectral indices between the maximum and minimum data points (maps)


  for (i = 1; i <=numregions; i++) {

    // Count how many times a regions has a flux below the rms
    if ( (regflux[map1][i] <= rms[map1])  || (regflux[map2][i] <= rms[map2]) ) {
      warningcount++;
      specindex[i] = 1e34; // These values will be ignored later 
      }

    else {
      if (freq1 > freq2) {
	specindex[i] = (-log(regflux[map2][i]/regflux[map1][i])) / (log(freq2/freq1));

	// Find the max and min spectral indices as we go
	if (specindex[i] > *maxsi) {
	  *maxsi = specindex[i];
	}
	if (specindex[i] < *minsi) {
	  *minsi = specindex[i];
	}
      }

      else if (freq2 > freq1) {
	specindex[i] = (-log(regflux[map1][i]/regflux[map2][i])) / (log(freq1/freq2));

	// Find the max and min spectral indices as we go
	if (specindex[i] > *maxsi) {
	  *maxsi = specindex[i];
	}
	if (specindex[i] < *minsi) {
	  *minsi = specindex[i];
	}
      }

      else {
	fprintf(stderr,"\nError: Cannot calculate spectral index. Are both maps the same frequency???\nExiting...\n\n");
	exit(0);
      }
    }


    if (printindex == 1) {

      printf("Spectal Index of region %d: %.2f\n", i, specindex[i]);
    }
  }

  /*
  if (warningcount > 0) {
    fprintf(stderr,"\n*** WARNING: Regions had a flux less than the RMS %d times! This normally occurs when fixed regions are used and they are not a good match ***\n\n", warningcount);
  }
  */

    printf("Maximum Spectal Index: %.2f\n", *maxsi);
    printf("Minimum Spectal Index: %.2f\n", *minsi);

  return 0;

}
