/* Simple module to plot spectral index against curvature

   What needs to be passed:

   float *curvature - Array of curvature values indexed in the same way as the spectral index array
   float *specindex - Array of spectral index values
   int regnumber - The total number of regions
   int symbol - Symbol to use for points (1 for dots). Se pgplot symbols for more options.
   float xmin, xmax, ymin, ymax - Minimum and maximum axis values

Basic usage:

  #include "plotcurvagainstspec.c"

  ...

  plotcurvagainstspec(cuvature, specindex, numberofregions, 1, 0, 10, 0 24);


   If you have any question please email jeremy.harwood@physics.org and I will try and answer them as soon as I can
*/

int plotcurvagainstspec(float **curvature, float *specindex, int regnumber, int symbol, float xmin, float xmax, float ymin, float ymax, int poly, int autoscalex, int autoscaley, float userminx, float usermaxx, float userminy, float usermaxy, char *settarget, int titles, int labels, int axismarks, int extwincontrol) {

  // Declaring variables
  int i;
  char maptitle[128];

  // Delcare pgpplot prototypes
  int cpgopen(const char *device);
  void cpgclos(void);
  void cpgenv(float xmin, float xmax, float ymin, float ymax, int just, int axis);
  void cpglab(const char *xlbl, const char *ylbl, const char *toplbl);
  void cpgpt1(float xpt, float ypt, int symbol);
  void cpgsci(int ci);
  void cpgscr(int ci, float cr, float cg, float cb);

  // Plot the spectral index map
  
  if (autoscalex == 1) {
    if (xmin < 0) {
      xmin *= 1.2;
    }
    else {
      xmin *= 0.8;
    }

    if (xmax < 0) {
      xmax *= 0.8;
    }
    else {
      xmax *= 1.2;
    }
  }
  else {
    xmin = userminx;
    xmax = usermaxx;
  }


  if (autoscaley == 1) {
    if (ymin < 0) {
      ymin *= 1.2;
    }
    else {
      ymin *= 0.8;
    }

    if (ymax < 0) {
      ymax *= 0.8;
    }
    else {
      ymax *= 1.2;
    }
  }
  else {

    ymin = userminy;
    ymax = usermaxy;
  }


  if (extwincontrol != 1) {
    cpgopen("/xs");
  }


  // Swap to background and forground colours to something sensible
  cpgscr(0, 1, 1, 1);
  cpgscr(1, 0, 0, 0);
 
  // Layout information
  cpgsci(1);

  if (axismarks == 1) {
    cpgenv(xmin,xmax,ymin,ymax, 0, 0);
  }
  else {
    cpgenv(xmin,xmax,ymin,ymax, 0, -1);
  }


 // Set plot title
  if (titles == 0) {
    sprintf(maptitle, " ");
  }
  else {
    sprintf(maptitle,"Spectral Index vs Spectral Curvature for %s", settarget);
  }
  
  if (labels == 0) {
    cpglab("", "", maptitle);
  }
  else {
    cpglab("Spectral Index", "Spectral Curvature", maptitle);
  }


  // Plotting each region

  for(i=0; i<regnumber; i++) {

    cpgpt1(specindex[i], curvature[i][poly], symbol);

  }

if (extwincontrol != 1) {
  cpgclos();
}

  return 0;

}




int plotmodelvsflux(float *modelflux, float *regflux, float *fluxerrorplus, float *fluxerrorminus, int symbol, int nummodelpts, int imgnum, float *fluxfrequency, float *modelfrequency, int autoscalex, int autoscaley, float userminx, float usermaxx, float userminy, float usermaxy, float minfreq, float maxfreq, float minflux, float maxflux, char *settarget, int titles, int labels, int axismarks, double inject, double fieldstrength, float bestage, float chisquared, int paramlabels, int model, int usereduced) {


  // Declaring variables

  double xaxismin, xaxismax, yaxismin, yaxismax;

  char str_inject[64], str_field[64], str_age[64], str_chi[64], ylabel[64], xlabel[64], maptitle[128];

  
  // Delcare pgpplot prototypes
  int cpgopen(const char *device);
  void cpgclos(void);
  void cpgenv(float xmin, float xmax, float ymin, float ymax, int just, int axis);
  void cpglab(const char *xlbl, const char *ylbl, const char *toplbl);
  void cpgpt(int n, const float *xpts, const float *ypts, int symbol);
  void cpgsci(int ci);
  void cpgerrb(int dir, int n, const float *x, const float *y, const float *e, float t);
  void cpgline(int n, const float *xpts, const float *ypts);
  void cpgerry(int n, const float *x, const float *y1, const float *y2, float t);
  void cpgscr(int ci, float cr, float cg, float cb);
  void cpgmtxt(const char *side, float disp, float coord, float fjust, const char *text);

  // Swap to background and forground colours to something sensible
  cpgscr(0, 1, 1, 1);
  cpgscr(1, 0, 0, 0);

  // Find the overall min and max values

  if (autoscalex == 1) {
    xaxismin = 0.9 * minfreq;
    xaxismax = 1.1 * maxfreq;
  }
  else {
    xaxismin = userminx;
    xaxismax = usermaxx;

  }

  if (autoscaley == 1) {
    yaxismin = 0.9 * minflux;
    yaxismax = 1.1 * maxflux;
  }
  else {
    yaxismin = userminy;
    yaxismax = usermaxy;
  }


  xaxismin = log10(xaxismin);
  xaxismax = log10(xaxismax);
  yaxismin = log10(yaxismin);
  yaxismax = log10(yaxismax);

  // Layout information

  cpgsci(1);

  if (titles == 0) { sprintf(maptitle," ");  }

  if (labels == 1) {
    sprintf(xlabel,"log(Frequency / Hz)");
    sprintf(ylabel,"log(Flux / Jy)");
  }
  else {
    sprintf(xlabel," ");
    sprintf(ylabel," ");
  }


  if (titles == 0) {
    sprintf(maptitle," "); 
  }
  else {
    sprintf(maptitle,"Model vs Observed Flux as a Function of Frequency");
  }

  if (axismarks == 1) {
    cpgenv(xaxismin, xaxismax, yaxismin, yaxismax, 0, 0);
  }
  else {
    cpgenv(xaxismin, xaxismax, yaxismin, yaxismax, 0, -1);
  }

  cpglab(xlabel, ylabel, maptitle);


  // Plot each region

  cpgpt(imgnum, fluxfrequency, regflux, symbol);
  cpgerry(imgnum, fluxfrequency, fluxerrorplus, fluxerrorminus, 0.5);


  cpgsci(2);
  cpgline(nummodelpts+1, modelfrequency, modelflux);

  cpgsci(1);
  if (paramlabels == 1) {

    if (model == 1) {
      cpgmtxt("B", -5.5, 0.05, 0.0, "JP Model");
    }
    if (model == 2) {
      cpgmtxt("B", -5.5, 0.05, 0.0, "KP Model");
    }
    if (model == 3) {
      cpgmtxt("B", -5.5, 0.05, 0.0, "Tribble Model (JP)");
    }
    else if (model == 4) {
      cpgmtxt("B", -5.5, 0.05, 0.0, "Tribble Model (KP)");
    }

    sprintf(str_inject, "Injection index: %.2f", inject);
    sprintf(str_field, "B-Field: %.2e T", fieldstrength);
    sprintf(str_age, "Spectral Age: %.2f", bestage);

    if (usereduced == 1) {
      sprintf(str_chi, "Reduced Chi-Squared: %.2f", chisquared);
    }
    else {
      sprintf(str_chi, "Chi-Squared: %.2f", chisquared);
    }

    cpgmtxt("B", -4.5 , 0.05, 0.0, str_inject);
    cpgmtxt("B", -3.5 , 0.05, 0.0, str_field);
    cpgmtxt("B", -2.5 , 0.05, 0.0, str_age);
    cpgmtxt("B", -1.5 , 0.05, 0.0, str_chi);
  }



  return 0;

}

