#include <stdio.h>
#include <math.h>
#include <gsl_integ.h>
#include <stdlib.h>
#include <constants.h>
#include <bessel.h>

/* synchrotron code comes from SoS */

#define MAXDATA 100
#define XF1 1.0e-4
#define XF2 0.22e+2
#define INFIN  1.0e2

#ifndef INTCO
#define INTCO (3*ECF)/(4*PI*M_EL*M_EL*M_EL*V_C*V_C*V_C*V_C)
#define EMCO (sqrt(3)*ECF*ECF*ECF)/(4.0*PI*EPS_0*V_C*M_EL)
#endif

static double sinalpha;
static double axf1, axf2;
static double *fx;
static int xls=100;

double fr[MAXDATA],em[MAXDATA],eme[MAXDATA];
double (*ne)(double);
double (*ng)(double);

double gmin,gmax,power,gbreak,gbvalue;
double EMIN,EMAX,EBREAK;
double n0_ext;
double BFIELD, nu;
double ebrd,gbrd;

double ageb,age;

gsl_integration_workspace *w1,*w2;

double ng_pow(double g) {

  if (g<gmin || g>gmax) return 0.0;
  else return pow(g,-power);
}

double ne_pow(double e) {

  /* ne(E) dE is the number of electrons with energy between E and E+dE */

  /* delta is the power-law index */
  /* normalise to 1 at some arbitrary energy */

  if ((e<EMIN) || (e>EMAX)) return 0.0;
  else return n0_ext*pow(e,-power);
}

double ne_age(double e) {

  /* as above, but with a finite age single-burst model */
  /* loss is the term that corrects for loss. */

  static double loss;

  if ((e<EMIN) || (e>EMAX)) return(0.0);
  else {
    loss=(4*THOMSON/(6*M_EL*M_EL*V_C*V_C*V_C*MU_0))*ageb*ageb*e*age;
    if (loss>=1.0) return(0.0);
    else return n0_ext*pow(e,-power)*pow((1.0-loss),power-2);
  }
}

double ng_age(double g) {

  /* as above, but with a finite age single-burst model */
  /* loss is the term that corrects for loss. */

  static double loss;

  if (g<gmin || g>gmax) return 0.0;
  else {
    loss=(4*THOMSON/(6*M_EL*V_C*MU_0))*ageb*ageb*g*age;
    if (loss>=1.0) return(0.0);
    else return pow(g,-power)*pow((1.0-loss),power-2);
  }
}

double age_emax(double limit) {

  double em;

  em=6.0*M_EL*M_EL*V_C*V_C*V_C*MU_0/(4*THOMSON*ageb*ageb*age);
  if (em>limit) em=limit;
  return(em);
}

double ng_break(double g) {
  if (g<gmin || g>gmax) return 0.0;
  else if (g>=gbreak) return gbrd*pow(g,-(power+gbvalue));
  else return pow(g,-power);
}

double ne_break(double e) {

  /* as above, but with a break */
  
  if ((e<EMIN) || (e>EMAX)) return 0.0;
  else if (e>=EBREAK) return n0_ext*ebrd*pow(e,-(power+gbvalue));
  else return n0_ext*pow(e,-power);
}

double ene(double e) {
  return e*ne(e);
}

double ffx(double x)

{
  /* use the tabulated values of F(x) to return F(x) at any x. For x
     outside the table, use the asymptote of Pacholczyk (1970) for small
     x and 0 for large x. Otherwise return a log-linear two-point
     interpolation. */

  int i;
  double av,xv1,xv2;

  if (x>=XF2)
    return 0.0;
  else if (x<=XF1)
    return 4.0*PI*pow((0.5*x),(1.0/3.0))/(sqrt(3.0)*2.68357);
  /* this is a cheat because I'm having problems with the gamma fn */
  else {
    av=log(x);
    xv1=(av-axf1)*((double)xls-1.0)/(axf2-axf1);
    xv2=floor(xv1);
    i=xv2;
    xv1-=xv2;
    return exp(((1.0-xv1)*log(fx[i]))+(xv1*log(fx[i+1])));
  }
}

double ff(double x) {

  float bessk, dummy1, dummy2, dummy3;

  /* given x, returns the Bessel function of order 5/3 */
  bessik(x,5.0/3.0,&dummy1,&bessk,&dummy2,&dummy3);
  /*  printf("x was %g, Bessel result was %g\n",x,bessk); */
  return bessk;
}

void makefx(void) {

  int i;
  double xv, yv, av;

  printf("Making F(x) lookup table:\n"); 
  axf1=log(XF1);
  axf2=log(XF2);
  for (i=0; i<xls; i++) {
    /* compute the x-value */
    av=axf1+((axf2-axf1)*((double)i)/(double)(xls-1));
    xv=exp(av);
    /* integrate ff from xv to infinity */
    yv=gsl_integ_iu(ff,xv,w1);
    //  printf("xv is %g, yv is %g\n",xv,yv);
    //  if (yv==FAILTAG) break;
    /* once qromo1 has failed to converge, assume that the rest are 0 */
    /* F(x) = x times integral */
    fx[i]=xv*yv;
    //  printf("%f %g\n",xv,fx[i]);
  }

  for (;i<xls; i++) fx[i]=0.0;
}

double integr(double e) {

  /* the synchrotron emission integrand */
  return ne(e)*ffx(nu/(INTCO*BFIELD*sinalpha*e*e));
}

double synchpl(void) {

  static double emin,emax,res;

  emin=sqrt(nu/(BFIELD*sinalpha*INTCO*XF2));
  emax=3*sqrt(nu/(BFIELD*sinalpha*INTCO*XF1));
  if (emin<EMIN) emin=EMIN;
  if (emax>EMAX) emax=EMAX;
  if (age>0.0) emax=age_emax(EMAX);
  else emax=EMAX;
  /* printf("Using min energy %l\n",emin); */
  /*  printf("in synchpl, emin = %g, emax = %g\n",emin,EMAX); */
  if (emin>emax) {
    /* printf("Synch. integration would be void; no suitable electrons.\n"); */
    res=0;
  } else
    res=gsl_integ(integr,emin,emax,w1);
  if (res<0) {
    printf("synchrotron integral convergence problem\n");
    res=0.0;
  }
  /*  printf("returning %g\n",res); */
  return EMCO*BFIELD*sinalpha*res;
}

double synchin(double alpha) {

  sinalpha=sin(alpha);
  return sinalpha*synchpl();
}

double synchint(void) {

  /* integrate over all pitch angles */

  return 0.5*gsl_integ(synchin,0.0,PI,w2);
}

double emiss_n(double n0, double b, double nu_arg) {

  /* similar to emiss_a, but the parameters of the electron
     distribution are fixed at init time, and can't be changed. */

  n0_ext=n0;
  BFIELD=b;
  nu=nu_arg;
  return synchint();
}
